#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <proto.h>

#include "gtappdisp.h"
#include "gtapp.h"
#include "gtwidget.h"
#include "gtconfig.h"
#include "gtcommon.h"
#include "gterror.h"

/************************************************************************\
*
* GtDisplay class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define R_DSPSTR		"dispstring"

/* Type of elements in hash table lists */
typedef struct GtHashList
{
	GtSLink lnk;
	Window win;
	GtWidget *obj;
} GtHashList;

/* Type of elements in selection list */
typedef struct GtSelList
{
	GtSLink lnk;
	GtWidget *obj;
	GtCallB Conv;
	GtCallB Lose;
	GtCallB Done;
	GtUDPtr ud;
	int transf;
	Atom sel;
	Time tm;
	uword flags;
} GtSelList;

#define GTSLF_OWNER		0x0001	/* Owner of selection */
#define GTSLF_INCR		0x0002	/* Incremental transfer */

/* Type of elements in requests list */
typedef struct GtReqList
{
	GtSLink lnk;
	GtSelList *sl;
	Window req;
	Atom sel;
	Atom target;
} GtReqList;	

/* Type of elements in selection get list */
typedef struct GtGetList
{
	GtSLink lnk;
	GtWidget *obj;
	Atom sel;
	Atom dest;
	GtCallB Got;
	GtUDPtr ud;
} GtGetList;	

/* Old error handler from safe ChangeProperty */
static int (*olderr)(Display *dsp,XErrorEvent *xee);

/************************************************************************\
*
* Ignore alloc error from Xlib.
*
\************************************************************************/
static int GtIgnoreAlloc(Display *dsp,XErrorEvent *xee)
{
	dsp=dsp;
	if(xee->error_code==BadAlloc)
		return(0);
	if(olderr)
		return(olderr(dsp,xee));
	return(0);
}

/************************************************************************\
*
* Create and init display.
*
\************************************************************************/
GtAppDisp *GtMkAppDisp(GtApp *a,GtTag *tgs)
{
	GtAppDisp *obj;
	
	if(!(obj=new GtAppDisp()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtAppDisp object!\n");
		return(NULL);
	}
	if(!obj->Init(a,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init app.
*
\************************************************************************/
GtAppDisp *GtMkAppDisp(GtApp *a,ulong tag,...)
{
	va_list va;
	GtAppDisp *obj;
	int p;
	
	va_start(va,tag);
	p=GtBeginVaTags();
	obj=GtMkAppDisp(a,GtCopyVaTags(tag,&va));
	GtEndVaTags(p);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty display.
*
\************************************************************************/
GtAppDisp::GtAppDisp(void)
{
	int n;
	
	GtSInit(&sels);
	GtSInit(&reqs);
	GtSInit(&gets);
	app=NULL;
	root=NULL;
	focus=NULL;
	grab=NULL;
	curr=NULL;
	dspdb=NULL;
	mclkt=GTDEF_MULTICLICK;
	for(n=0; n<GTD_HASH_SIZE; n++)
		GtSInit(&hash[n]);
}

/************************************************************************\
*
* Destroy display.
*
\************************************************************************/
void GtAppDisp::Destroy(void)
{
	if(app)
		app->RemDisp(this);
	if(root)
		GtDestroy(root);
	if(dspdb)
		XrmDestroyDatabase(dspdb);
	GtDisplay::Destroy();
}

/************************************************************************\
*
* Setup display.
*
\************************************************************************/
sbool GtAppDisp::Init(GtApp *a,GtTag *tgs)
{
	GtTag *t,tg[2];
	char *dname=NULL;
	char *dbstr;
	XrmDatabase tmp=NULL;
	sbool mc=FALSE;
	
	app=a;
	if(app)
		app->GetRes(NULL,R_DSPSTR,GTRT_STRING,&dname);
	tg[0].tag=GT_DispString;
	tg[0].d.ptr=dname;
	tg[1].tag=GT_MORE;
	tg[1].d.ptr=tgs;
	if(!GtDisplay::Init(tg))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_MultiClickTime:
				mclkt=t->d.val;
				mc=TRUE;
				break;
		}
	}
	if(!(stdat[GTDSA_TARGETS]=XInternAtom(dsp,"TARGETS",False)))
	{
		GtError(this,GTEC_NOXRESOURCE,"No targets atom!\n");
		return(FALSE);
	}
	if(!(stdat[GTDSA_TIMESTAMP]=XInternAtom(dsp,"TIMESTAMP",False)))
	{
		GtError(this,GTEC_NOXRESOURCE,"No timestamp atom!\n");
		return(FALSE);
	}
	if(!(stdat[GTDSA_MULTIPLE]=XInternAtom(dsp,"MULTIPLE",False)))
	{
		GtError(this,GTEC_NOXRESOURCE,"No multiple atom!\n");
		return(FALSE);
	}
	if((dbstr=XResourceManagerString(dsp)))
	{
		dspdb=XrmGetStringDatabase(dbstr);
		if((dbstr=XScreenResourceString(DefaultScreenOfDisplay(dsp))) &&
				(tmp=XrmGetStringDatabase(dbstr)))
			XrmMergeDatabases(tmp,&dspdb);
	}
	if(app && !mc)
		app->GetRes(NULL,"multiclicktime",GTRT_LONG,&mclkt);
	if(app)
		app->AddDisp(this);
	return(TRUE);
}

/************************************************************************\
*
* Find app tag and then setup normal.
*
\************************************************************************/
sbool GtAppDisp::Init(GtTag *tgs)
{
	GtApp *a;
	
	a=(GtApp *)GtGetTagPtr(tgs,GT_App,NULL);
	return(this->Init(a,tgs));
}

/************************************************************************\
*
* Set display attributes.
*
\************************************************************************/
sbool GtAppDisp::Set(GtTag *tgs)
{
	GtTag *t;
	
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_MultiClickTime:
				mclkt=t->d.val;
				break;
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Get display attributes.
*
\************************************************************************/
void GtAppDisp::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_App:
				*ptr=(long)app;
				break;
			case GT_MultiClickTime:
				*ptr=mclkt;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is a (sub)class of named class.
*
\************************************************************************/
sbool GtAppDisp::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtAppDisp")==0)
		return(TRUE);
	else
		return(GtBase::IsSubClassOf(cn));
}

/************************************************************************\
*
* Handle one X-event.
*
\************************************************************************/
void GtAppDisp::HandleEvent(XEvent *xe)
{
	XEvent xetmp;
	GtHashList *hl;
	GtWidget *obj=NULL,*tmp;
	sbool okgrab=TRUE;
	
	if(!xe)
		xe=&xetmp;
	GtDisplay::HandleEvent(xe);
	if(xe->type==SelectionRequest)
		HandleSelReq(xe);
	else if(xe->type==SelectionClear)
		HandleSelClr(xe);
	else if((xe->type==PropertyNotify) && (xe->xproperty.state==PropertyDelete))
		HandlePropDel(xe);
	else if(xe->type==SelectionNotify)
		HandleSelGet(xe);
	if(precb.CallCB(this,xe)==GTCB_BREAK)
		return;
	if(focus && (focus->Handle(xe)!=GTHR_SEARCH))
		return;
	if(root)
	{
		for(hl=(GtHashList *)GtSFirst(&hash[xe->xany.window % GTD_HASH_SIZE]); hl;
				hl=(GtHashList *)GtSNext(&hl->lnk))
		{
			if(hl->win==xe->xany.window)
			{
				obj=hl->obj;
				break;
			}
		}
		if(!obj)
			return;
		if(grab)
		{
			if(!obj->IsChildOf(grab) && (xe->type!=Expose))
			{
				curr=grab;
				tmp=grab;
				grab->Handle(xe);
				curr=NULL;
				if(tmp->DoDestroy())
					GtDestroy(tmp);
				return;
			}
		}
		tmp=obj;
		while(obj)
		{
			curr=obj;
			if(obj->Handle(xe)==GTHR_USED)
				break;
			if(grab==obj)
				okgrab=FALSE;
			obj=(GtWidget *)obj->GetParent();
		}
		curr=NULL;
		for(obj=tmp; obj; obj=(GtWidget *)obj->GetParent())
		{
			if(obj->DoDestroy())
				GtDestroy(obj);
		}
		if(grab && okgrab)
		{
			curr=grab;
			tmp=grab;
			grab->Handle(xe);
			curr=NULL;
			if(tmp->DoDestroy())
				GtDestroy(tmp);
		}
	}
	return;
}

/************************************************************************\
*
* Add widget to hash list.
*
\************************************************************************/
void GtAppDisp::AddWidget(GtWidget *w)
{
	GtHashList *hl;
	
	if(!w || w->IsWindowLess())
		return;
	if(!(hl=(GtHashList *)malloc(sizeof(GtHashList))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for hash list!\n");
		return;
	}
	hl->obj=w;
	hl->win=w->GetWin()->XWin();
	GtSAddTail(&hash[hl->win % GTD_HASH_SIZE],&hl->lnk);
}

/************************************************************************\
*
* Remove widget from hash list.
*
\************************************************************************/
void GtAppDisp::RemWidget(GtWidget *w)
{
	GtHashList *hl;
	GtSelList *sl;
	GtReqList *rl;
	GtGetList *gl;
	Window win;
	
	if(!w || w->IsWindowLess())
		return;
	win=w->GetWin()->XWin();
	for(hl=(GtHashList *)GtSFirst(&hash[win % GTD_HASH_SIZE]); hl;
			hl=(GtHashList *)GtSNext(&hl->lnk))
	{
		if(hl->obj==w)
		{
			GtSRemove(&hash[win % GTD_HASH_SIZE],&hl->lnk);
			free(hl);
			break;
		}
	}
	for(sl=(GtSelList *)GtSFirst(&sels); sl;
			sl=(GtSelList *)GtSNext(&sl->lnk))
	{
		if(sl->obj==w)
		{
			GtSRemove(&sels,&sl->lnk);
			break;
		}
	}
	for(rl=(GtReqList *)GtSFirst(&reqs); rl;
			rl=(GtReqList *)GtSNext(&rl->lnk))
	{
		if(rl->sl==sl)
		{
			GtSRemove(&reqs,&rl->lnk);
			free(rl);
			break;
		}
	}
	if(sl)
		free(sl);
	for(gl=(GtGetList *)GtSFirst(&gets); gl;
			gl=(GtGetList *)GtSNext(&gl->lnk))
	{
		if(gl->obj==w)
		{
			GtSRemove(&gets,&gl->lnk);
			free(gl);
			break;
		}
	}
}

/************************************************************************\
*
* Set focus widget.
*
\************************************************************************/
void GtAppDisp::SetFocus(GtWidget *o)
{
	if(!focus)
	{
		if(grab && !o->IsChildOf(grab))
			return;
		focus=o;
		focus->Notify(GT_FocusIn);
	}
}

/************************************************************************\
*
* Clear focus widget.
*
\************************************************************************/
void GtAppDisp::ClrFocus(GtWidget *o)
{
	if(focus && (focus==o))
	{
		focus->Notify(GT_FocusOut);
		focus=NULL;
	}
}

/************************************************************************\
*
* Check if child of an object is in focus and if so clear focus.
*
\************************************************************************/
void GtAppDisp::ChkFocus(GtWidget *o)
{
	if(focus && focus->IsChildOf(o))
	{
		focus->Notify(GT_FocusOut);
		focus=NULL;
	}
}
		
/************************************************************************\
*
* Set grab widget.
*
\************************************************************************/
void GtAppDisp::SetGrab(GtWidget *o)
{
	if(!grab)
	{
		grab=o;
		if(focus && !focus->IsChildOf(grab))
		{
			focus->Notify(GT_FocusOut);
			focus=NULL;
		}
	}
}

/************************************************************************\
*
* Clear grab widget.
*
\************************************************************************/
void GtAppDisp::ClrGrab(GtWidget *o)
{
	if(grab==o)
		grab=NULL;
}

/************************************************************************\
*
* Set selection owner to given widget.
*
\************************************************************************/
sbool GtAppDisp::OwnSelection(GtWidget *w,Atom s,Time t,
	GtCallB conv,GtCallB lose,GtCallB done,GtUDPtr ud)
{
	GtSelList *sl;
	
	if(w->GetWin())
	{
		if(!(sl=(GtSelList *)malloc(sizeof(GtSelList))))
			return(FALSE);
		XSetSelectionOwner(dsp,s,w->GetWin()->XWin(),t);
		if(XGetSelectionOwner(dsp,s)==w->GetWin()->XWin())
		{
			GtSAddHead(&sels,&sl->lnk);
			sl->obj=w;
			sl->Conv=conv;
			sl->Lose=lose;
			sl->Done=done;
			sl->ud=ud;
			sl->transf=0;
			sl->sel=s;
			sl->tm=t;
			sl->flags=GTSLF_OWNER;
			return(TRUE);
		}
		else
			free(sl);
	}
	return(FALSE);
}

/************************************************************************\
*
* Release selection from given widget.
*
\************************************************************************/
void GtAppDisp::DisOwnSelection(GtWidget *w,Atom sel)
{
	GtSelList *sl;
	GtSelClrCD cd;
	
	for(sl=(GtSelList *)GtSFirst(&sels); sl;
			sl=(GtSelList *)GtSNext(&sl->lnk))
	{
		if((w==sl->obj) && (sel==sl->sel))
		{
			XSetSelectionOwner(dsp,sl->sel,None,sl->tm);
			sl->flags&=~GTSLF_OWNER;
			if(!sl->transf)
			{
				cd.xevent=NULL;
				cd.w=sl->obj;
				cd.sel=None;
				if(sl->Done)
					sl->Done(this,&cd,sl->ud);
				GtSRemove(&sels,&sl->lnk);
				free(sl);
			}
		}
	}
}

/************************************************************************\
*
* Get selection value.
*
\************************************************************************/
void GtAppDisp::GetSelection(GtWidget *w,Atom sel,Atom target,Atom dest,
															Time tm,GtCallB cb,GtUDPtr ud)
{
	GtGetList *gl;
	
	if(!(gl=(GtGetList *)malloc(sizeof(GtGetList))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for get selection!\n");
		return;
	}
	gl->obj=w;
	gl->sel=sel;
	gl->dest=dest;
	gl->Got=cb;
	gl->ud=ud;
	GtSAddHead(&gets,&gl->lnk);
	XConvertSelection(dsp,sel,target,dest,w->GetWin()->XWin(),tm);
}

/************************************************************************\
*
* Add callback function.
*
\************************************************************************/
void GtAppDisp::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	switch(tag)
	{
		case GT_Before:
			precb.AddCB(cb,ud);
			break;
		case GT_EnterLeave:
			entercb.AddCB(cb,ud);
			break;
	}
}

/************************************************************************\
*
* Remove callback function.
*
\************************************************************************/
void GtAppDisp::RemCB(ulong tag,GtCallB cb)
{
	switch(tag)
	{
		case GT_Before:
			precb.RemCB(cb);
			break;
		case GT_EnterLeave:
			entercb.RemCB(cb);
			break;
	}
}

/************************************************************************\
*
* Activate enter/leave callback for a widget.
*
\************************************************************************/
void GtAppDisp::EnterLeave(GtWidget *w,sbool enter,ulong state)
{
	GtEnterLeaveCD cd;
	
	cd.wid=w;
	cd.state=state;
	cd.enter=enter;
	entercb.CallCB(this,&cd);
}

/************************************************************************\
*
* Handle selection requests.
*
\************************************************************************/
void GtAppDisp::HandleSelReq(XEvent *xe)
{
	GtSelList *sl;
	GtSelReqCD cd;
	XEvent rxe;
	Window req;
	Atom target,prop,type;
	int fmt,res;
	ulong len,blen,n;
	Atom *data;
	
	req=xe->xselectionrequest.requestor;
	target=xe->xselectionrequest.target;
	prop=xe->xselectionrequest.property;
	rxe.type=SelectionNotify;
	rxe.xselection.requestor=req;
	rxe.xselection.selection=xe->xselectionrequest.selection;
	rxe.xselection.target=target;
	rxe.xselection.property=None;
	rxe.xselection.time=xe->xselectionrequest.time;
	for(sl=(GtSelList *)GtSFirst(&sels); sl;
			sl=(GtSelList *)GtSNext(&sl->lnk))
	{
		if((sl->flags & GTSLF_OWNER) &&
				(xe->xselectionrequest.owner==sl->obj->GetWin()->XWin()) && 
				(xe->xselectionrequest.selection==sl->sel))
			break;
	}
	if(!sl || ((prop==None) && (target==stdat[GTDSA_MULTIPLE])))
	{
		XSendEvent(dsp,req,True,0,&rxe);
		return;
	}
	if(prop==None)
		prop=target;
	if(target==stdat[GTDSA_TIMESTAMP])
	{
		SafeChange(req,prop,target,32,PropModeReplace,(ubyte *)&sl->tm,1);
		if(AddSelReq(sl,req,prop,target))
		{
			rxe.xselection.property=prop;
			sl->transf++;
		}
		XSendEvent(dsp,req,True,0,&rxe);
	}
	else if(target==stdat[GTDSA_MULTIPLE])
	{
		/* This code may not be correct */
		// Get the property with target/property pairs.
		res=XGetWindowProperty(dsp,req,prop,0,0,False,
			AnyPropertyType,&type,&fmt,&len,&blen,(ubyte **)&data);
		if(res==Success)
		{
			XFree(cd.data);
			res=XGetWindowProperty(dsp,req,prop,0,
				blen/4+1,False,AnyPropertyType,&type,&fmt,&len,&blen,
				(ubyte **)&data);
		}
		// Unable to get property.
		if(res!=Success)
		{
			XSendEvent(dsp,req,True,0,&rxe);
			return;
		}
		// Loop through all pairs.
		for(n = 0; n < (len/2); n++)
		{
			if(data[2*n+1] == None)
				continue;
			cd.xevent=xe;
			cd.w=sl->obj;
			cd.req=req;
			cd.sel=xe->xselectionrequest.selection;
			cd.target=data[2*n];
			cd.prop=data[2*n+1];
			cd.fmt=8;
			cd.data=NULL;
			cd.len=0;
			cd.ok=TRUE;
			// Convert a pair.
			sl->Conv(this,&cd,sl->ud);
			// Write property or set pair to indicate failure.
			if(cd.ok)
				SafeChange(req,data[2*n+1],cd.target,cd.fmt,PropModeAppend,cd.data,cd.len);
			else
				data[2*n] = None;
		}
		// Replace property with pairs.
		SafeChange(req,prop,target,fmt,PropModeReplace,(ubyte *)data,len);
		if(AddSelReq(sl,req,prop,target))
		{
			rxe.xselection.property=prop;
			sl->transf++;
		}
		XSendEvent(dsp,req,True,0,&rxe);
	}
	else
	{
		cd.xevent=xe;
		cd.w=sl->obj;
		cd.req=req;
		cd.sel=xe->xselectionrequest.selection;
		cd.target=target;
		cd.prop=prop;
		cd.fmt=8;
		cd.data=NULL;
		cd.len=0;
		cd.ok=TRUE;
		sl->Conv(this,&cd,sl->ud);
		if(cd.ok && AddSelReq(sl,req,prop,target))
		{
			SafeChange(req,prop,cd.target,cd.fmt,PropModeAppend,cd.data,cd.len);
			rxe.xselection.property=prop;
			sl->transf++;
		}
		XSendEvent(dsp,req,True,0,&rxe);
	}
}

/************************************************************************\
*
* Handle selection clear messages.
*
\************************************************************************/
void GtAppDisp::HandleSelClr(XEvent *xe)
{
	GtSelList *sl;
	GtSelClrCD cd;
	
	for(sl=(GtSelList *)GtSFirst(&sels); sl;
			sl=(GtSelList *)GtSNext(&sl->lnk))
	{
		if((sl->flags & GTSLF_OWNER) &&
				(xe->xselectionclear.window==sl->obj->GetWin()->XWin()) && 
				(xe->xselectionclear.selection==sl->sel))
		{
			cd.xevent=xe;
			cd.w=sl->obj;
			cd.sel=xe->xselectionclear.selection;
			if(sl->Lose)
				sl->Lose(this,&cd,sl->ud);
			sl->flags&=~GTSLF_OWNER;
			if(!sl->transf)
			{
				cd.xevent=xe;
				cd.w=sl->obj;
				cd.sel=xe->xselectionclear.selection;
				if(sl->Done)
					sl->Done(this,&cd,sl->ud);
				GtSRemove(&sels,&sl->lnk);
				free(sl);
			}
			break;
		}
	}
}

/************************************************************************\
*
* Handle selection notify.
*
\************************************************************************/
void GtAppDisp::HandleSelGet(XEvent *xe)
{
	GtGetList *gl;
	GtSelGetCD cd;
	ulong blen;
	int res;
	
	for(gl=(GtGetList *)GtSFirst(&gets); gl;
			gl=(GtGetList *)GtSNext(&gl->lnk))
	{
		if((xe->xselection.requestor==gl->obj->GetWin()->XWin()) &&
				(xe->xselection.selection==gl->sel) &&
				(xe->xselection.property==gl->dest))
		{
			break;
		}
	}
	if(gl)
	{
		cd.w=gl->obj;
		cd.sel=gl->sel;
		res=XGetWindowProperty(dsp,gl->obj->GetWin()->XWin(),gl->dest,0,0,False,
			AnyPropertyType,&cd.type,&cd.fmt,&cd.len,&blen,(ubyte **)&cd.data);
		if(res==Success)
		{
			XFree(cd.data);
			res=XGetWindowProperty(dsp,gl->obj->GetWin()->XWin(),gl->dest,0,
				blen/4+1,True,AnyPropertyType,&cd.type,&cd.fmt,&cd.len,&blen,
				(ubyte **)&cd.data);
		}
		if(res!=Success)
		{
			cd.type=None;
			cd.fmt=0;
			cd.data=NULL;
			cd.len=0;
		}
		gl->Got(this,&cd,gl->ud);
		if(cd.data)
			XFree(cd.data);
		GtSRemove(&gets,&gl->lnk);
		free(gl);
	}
}

/************************************************************************\
*
* Handle property delete on selection windows.
*
\************************************************************************/
void GtAppDisp::HandlePropDel(XEvent *xe)
{
	GtReqList *rl;
	GtSelClrCD cd;
	
	for(rl=(GtReqList *)GtSFirst(&reqs); rl;
			rl=(GtReqList *)GtSNext(&rl->lnk))
	{
		if((rl->req==xe->xproperty.window) && (rl->target==xe->xproperty.atom))
		{
			rl->sl->transf--;
			if(!rl->sl->transf && !(rl->sl->flags & GTSLF_OWNER))
			{
				cd.xevent=xe;
				cd.w=rl->sl->obj;
				cd.sel=rl->sel;
				if(rl->sl->Done)
					rl->sl->Done(this,&cd,rl->sl->ud);
				GtSRemove(&sels,&rl->sl->lnk);
				free(rl->sl);
			}
			GtSRemove(&reqs,&rl->lnk);
			free(rl);
			break;
		}
	}
}

	
/************************************************************************\
*
* Add selection request to list.
*
\************************************************************************/
sbool GtAppDisp::AddSelReq(GtSelList *sl,Window req,Atom prop,Atom target)
{
	GtReqList *rl;
	
	if(!(rl=(GtReqList *)malloc(sizeof(GtReqList))))
		return(FALSE);
	rl->sl=sl;
	rl->req=req;
	rl->sel=prop;
	rl->target=target;
	GtSAddHead(&reqs,&rl->lnk);
	return(TRUE);
}

/************************************************************************\
*
* Make a safe ChangeProperty for selection handling.
*
\************************************************************************/
void GtAppDisp::SafeChange(Window req,Atom prop,Atom target,int fmt,int mode,
														ubyte *data,int len)
{
	olderr=XSetErrorHandler(GtIgnoreAlloc);
	XChangeProperty(dsp,req,prop,target,fmt,mode,data,len);
	XSetErrorHandler(olderr);
}
